"""
This script applies a patch to log4j to fix the CVE-2021-44228
vulnerability.

It solves it by removing the JndiLookup.class from the
log4j library that contains the vulnerability.

Usage:

1. Start the limefu command prompt.
2. Make sure that the Lime Python environment is active.
3. Run the following:

```
$ python patch-log4j-CVE-2021-44228.py
```
"""
import pathlib
import zipfile
import os.path
import os
import shutil
import click
import win32serviceutil
import win32service
import sys
import time
import ctypes


ELASTIC_SVC_NAME = 'limepro-elasticsearch'


DEFAULT_LIME_INSTALL_DIR = pathlib.Path(
    os.path.expandvars(r'%programfiles(x86)%\Lundalogik')
)

DEFAULT_INSTALL_DIR = DEFAULT_LIME_INSTALL_DIR / 'LIME Pro Server'


FILE_TO_REMOVE = 'org/apache/logging/log4j/core/lookup/JndiLookup.class'


def _get_log4j_path(installdir):
    return (
        installdir / 'elasticsearch-5.0.0' /
        'lib' / 'log4j-core-2.6.2.jar'
    )


def _get_log4j_temp_path(installdir):
    return (
        installdir / 'elasticsearch-5.0.0' /
        'lib' / 'log4j-core-2.6.2-temp.jar'
    )


def _backup(installdir):
    timestamp = time.time()
    output_filename = installdir / 'elastic-backup-{}'.format(timestamp)
    click.secho('Backing up files to {}.zip'.format(output_filename))
    shutil.make_archive(
        output_filename, 'zip',
        installdir / 'elasticsearch-5.0.0'
    )


def _is_admin():
    try:
        return ctypes.windll.shell32.IsUserAnAdmin() != 0
    except Exception:
        return False


def _start_elastic():
    click.secho('Starting the Lime CRM Search Engine')
    try:
        win32serviceutil.StartService(ELASTIC_SVC_NAME)
        win32serviceutil.WaitForServiceStatus(
            ELASTIC_SVC_NAME, win32service.SERVICE_RUNNING, 60
        )
    except Exception:
        click.err('Failed to start service', err=True)
        sys.exit(1)


def _stop_elastic():
    click.secho('Stopping the Lime CRM Search Engine')
    if _elastic_is_stopped():
        return
    try:
        win32serviceutil.StopService(ELASTIC_SVC_NAME)
        win32serviceutil.WaitForServiceStatus(
            ELASTIC_SVC_NAME, win32service.SERVICE_STOPPED, 60
        )
    except Exception:
        click.secho('Failed to stop service', err=True)
        sys.exit(1)

    # Ensure that the service has time for releasing potential file
    # locks on the jar file we want to edit. Without this sleep
    # this command sometimes failed.
    time.sleep(5)


def _elastic_is_stopped():
    _, state, _, _, _, _, _ = win32serviceutil.QueryServiceStatus(
        ELASTIC_SVC_NAME
    )
    return state == win32service.SERVICE_STOPPED


def _elastic_is_hanging():
    _, state, _, _, _, _, _ = win32serviceutil.QueryServiceStatus(
        ELASTIC_SVC_NAME
    )
    return state not in [
        win32service.SERVICE_RUNNING,
        win32service.SERVICE_STOPPED,
    ]


def _patch_log4j(log4j_path, log4j_temp_path):
    if os.path.exists(log4j_temp_path):
        os.remove(log4j_temp_path)
    shutil.move(log4j_path, log4j_temp_path)

    zin = zipfile.ZipFile(log4j_temp_path, 'r')
    zout = zipfile.ZipFile(log4j_path, 'w')

    for item in zin.infolist():
        if item.filename == FILE_TO_REMOVE:
            print('Removed vulnerability')
            continue

        buffer = zin.read(item.filename)
        zout.writestr(item, buffer)

    zout.close()
    zin.close()

    os.remove(log4j_temp_path)


@click.command()
@click.option(
    '--installdir',
    default=DEFAULT_INSTALL_DIR,
    help=(
        'Installation path for Lime CRM Server '
        '(default is "{}")'.format(DEFAULT_INSTALL_DIR)
    )
)
@click.option('--prompt/--no-prompt', default=True)
def patch_log4j_cli(installdir, prompt):
    log4j_path = _get_log4j_path(installdir=installdir)
    log4j_temp_path = _get_log4j_temp_path(installdir=installdir)
    admin_message = (
        'This script needs to be run with administrative privileges. '
        'Please make sure to run this in a commpand prompt as administrator.'
    )
    prompt_message = (
        'This will apply a patch to log4j to fix the CVE-2021-44228 '
        'vulnerability.'
    )

    if not _is_admin():
        click.secho(admin_message, err=True)
        sys.exit(1)

    if prompt and not click.confirm(prompt_message):
        click.secho(
            'Program exited. Patch has not been applied!',
            fg='yellow'
        )
        sys.exit(1)

    if _elastic_is_hanging():
        click.secho(
            'The Lime CRM Search Engine service is in mode '
            '"STOPPING" or "STARTING" which means that the '
            'patch cannot be applied. Restart the server and '
            're-run this script.',
            fg='yellow'
        )
        sys.exit(1)

    _stop_elastic()
    _backup(installdir=installdir)
    _patch_log4j(log4j_path, log4j_temp_path)
    _start_elastic()
    click.secho('Patch has been applied', fg='green')


if __name__ == '__main__':
    patch_log4j_cli()

